def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        output_dir='Output',
        output_files=['root_output.txt',
                      'shoot_output.txt'],
        skip_check_results=True)
